-- Programatica Front-End Commands, level 0
module Pfe0Cmds where
import Prelude hiding (print,putStr,putStrLn,catch,readFile)
import List((\\))
import Monad(unless,when)
import Maybe(fromJust)

import PFE0

import HsName(ModuleName(..),isMainModule,sameModuleName)
--import HsConstants(main_mod)
import HsLexerPass1(rmSpace,Pos(line)) -- for lines of code metrics

import CmdLineParser
import PfeParse
import PrettyPrint
import MUtils
import Graph(reverseGraph,reachable)
import DirUtils(optCreateDirectory)
import AbstractIO
import Statistics

pfe0 parseModule args = runPFE0Cmds () pfe0Cmds parseModule args

pfe0Cmds = projman++graphqueries

addHelpCmd prg cmds0 = cmds where cmds=cmds0++helpCmd prg cmds

helpCmd prg cmds =
  [("help",(Null (putStrLn (usage prg cmds)),"list available commands"))]

projman =
  -- Project management (a project is a set of source files)
  [("new"     , (fileArgs   new, "create a new project containing (zero or more) <files>")),
   ("add"     , (fileArgs   add, "add files to the project")),
   ("remove"  , (fileArgs   remove, "remove files from the project")),
   ("prune"   , (moduleArgs prune,"remove unreachable modules from the project")),
   ("files"   , (Null       files,"list files in the project")),
   ("options" , (Null       options,"show options in effect"))]

graphqueries =
  -- Module graph queries
  [("modules" , (moduleArgs modules,"show a topologically sorted list of modules")),
   ("graph"   , (moduleArgs graph,"show module dependecy (sub)graph")),
   ("dotgraph", (moduleArgs dotgraph,"dot format module dependency graph")),
   ("revgraph", (moduleArgs revgraph,"show reverse module dependecy (sub)graph")),
   ("dirgraph", (moduleArgs dirgraph,"show source directory dependecy (sub)graph")),
   ("dotdirgraph", (moduleArgs dotdirgraph,"dot format source directory dependecy (sub)graph")),
   ("unused"  , (moduleArgs unused,"list unimported and unreachable modules")),
   ("file"    , (moduleArg  file,"which file is the module in")),
   ("module"  , (fileArg    module',"which module does the file contain")),
   ("loc"     , (moduleArgs locModules, "number of lines of code")),
   ("sizemetrics" , (Null sizeMetrics, "number of lines per module metrics")),
   ("locmetrics" , (Null locMetrics, "number of lines of code per module metrics")),
   ("importmetrics",(Null importMetrics, "number of imports per module metrics")),
   ("exportmetrics",(Null exportMetrics, "number of exports (importers) per module metrics"))]

--- Project management ---------------------------------------------------------
new args    = do newProject; addPaths args
add args    = addPaths args
remove args = removePaths args

files = putStr . unlines =<< allFiles

options = print =<< parserFlags

--- Module graph queries -------------------------------------------------------
modules ms = do g <- getSortedSubGraph ms
	        let sccs = map (map (fst.snd)) g
	        putStrLn "Topologically sorted strongly connected components:"
	        putStr (unlines (map (unwords.map pp) sccs))

graph = showGraph id
revgraph = showGraph reverseGraph
showGraph f = pput . makeFormat . f . map snd @@ getSubGraph 

getSubGraph ms = concat # getSortedSubGraph ms
getSortedSubGraph ms = flip optSubGraph (just ms) # sortCurrentModuleGraph

-- Dot format functions contributed by Claus Reinke:
dotgraph = showDotGraph id
showDotGraph f = pput . dotFormat . f . map snd @@ getSubGraph
     
dotFormat g =
   "digraph ModuleGraph {size=\"7.5,10.5\";ratio=fill;" $$
   vcat [q x<>"->"<>braces (fsep [q d<>";"|d<-xs])|(x,xs)<-g] $$
   "}"
  where
    q = doubleQuotes

makeFormat g = vcat [x<>":"<+>fsep xs|(x,xs)<-g]

dirgraph    ms = pput . makeFormat =<< srcdirGraph ms
dotdirgraph ms = pput . dotFormat =<< srcdirGraph ms

srcdirGraph ms =
  do g <- getSubGraph ms
     let moddirs = [(m,dirname path)|(path,(m,_))<-g]
	 moddir m = fromJust (lookup m moddirs)
         dg = mapSnd usort $
              collectByFst
                [(moddir m,dir')|(_,(m,ms))<-g,dir'<-usort (map moddir ms)]
     return dg

dirname = current . reverse . dropWhile (/='/') . reverse
  where current "" = "./"
	current path = path
     
file m =
    do g <- getCurrentModuleGraph
       pput $ vcat [m<>":"<+>f|(f,(m',_))<-g,m'==m]

module' f =
    do g <- getCurrentModuleGraph
       pput $ vcat [f<>":"<+>m|(f',(m,_))<-g,f'==f]

getUnused ms =
  do fg <- getCurrentModuleGraph
     let g = map snd fg
         unimported = map fst g \\ [m|(m,_:_)<-reverseGraph g]
	 roots = if null ms then mains else ms
	   where mains = [m|(m,_)<-g,isMainModule m]
	 r = reachable g roots
	 unreached = map fst g \\ r
     when (null roots) $
       fail "No root modules given, no Main module in the project"
     return (fg,roots,unimported,unreached)

mainModules =
  do fg <- getCurrentModuleGraph
     return [m|(f,(m,_))<-fg,isMainModule m]

unused ms =
  do (_,roots,us,unr) <- getUnused ms
     if null us -- unlikely...
       then putStrLn "All modules are imported somewhere"
       else pput $ "The following modules are never imported:" $$
	           nest 4 (fsep us)
     unless (null unr) $
       do pput $
	    sep [ppi "The following modules are unreachable from",
		 nest 2 (fsep roots<>":")]
	  pput $ nest 4 (fsep unr)

prune ms =
  do (fg,_,_,unreached) <- getUnused ms
     remove [f|(f,(m,_))<-fg,m `elem` unreached]

sizeMetrics =
    pput . ppStatistics "number of lines" "module"
      =<< mapM (forkM (fst.snd) (size #. readFile . fst))
      =<< getCurrentModuleGraph
  where
    size = length . lines

locMetrics =
    pput . ppStatistics "number of lines of code" "module"
      =<< mapM (forkM (fst.snd) (loc #. lex0SourceFile.fst))
      =<< getCurrentModuleGraph

--forkM f g x = (,) (f x) # g x
forkM f g x =
  do y <- g x
      -- seq to process one file at a time in locMetrics & sizeMetrics
     seq y (return (f x,y))

loc = length . squeezeDups . map (line.fst.snd) . rmSpace

locModules = mapM_ locModule
locModule = locFile @@ findFile
locFile = pput . loc @@ lex0SourceFile


importMetrics = graphMetrics id "number of imports"
exportMetrics = graphMetrics reverseGraph "number of importers"

graphMetrics f lbl =
    do g <- mapSnd length . f . map snd # getCurrentModuleGraph
       pput $ ppStatistics lbl "module" g
